Long running tests that are to be run manually and not on CRAN.
They should be run routinely to ensure sanity of CVXR, even if
not on CRAN.


##library(testthat)
##library(CVXR)
##Sys.setenv(NOT_CRAN = TRUE)
##for (f in list.files(".", pattern = ".R$")) test_file(f)
